#'Plot maps in Robinson projection or other
#'@description Transforms a regular grid to a new projection (Robinson) and display it
#'@param data an array (only draws a map each time)
#'@param original_proj the data projection by default long-lat regular WGS 84               
#'@param target_proj the desired projection by default Robinson 
#'
#'@param legend possibilities are NULL (no color bar), 's2dv' adds ColorBar, 'ggplot2' 
#'              the result from ggplot2 and 'test' that makes a filled scale
#'@param ... extra arguments for ColorBar from s2dv
#'@import sf ggplot2 rnaturalearth cowplot
PlotRobinsonMap <- function(data, lon, lat, original_proj = 4326, 
                            target_proj = 54030, brks, cols, colNA = 'green',
                            size = 0.05, coastlines_width = 0.3,
                            dots = NULL, dots_shape = 47, dots_size = 0.5,
                            toptitle = NULL, caption = NULL,
                            fileout = NULL, legend = NULL,
                            lon_dim = 'lon', lat_dim = 'lat', 
                            width = 5, height = 4, dpi = 300, units = "in", 
                            device = 'png', triangle_ends = NULL,
                            col_inf = NULL, col_sup = NULL, vertical = TRUE,
                            ...) {
  data <- drop(data)
  dims <- dim(data)
  if (length(dims) > 2) {
    stop("Parameter 'data' has more than two dimension.")
  }
  if (!all(names(dims) %in% c(lat_dim, lon_dim))) {
    stop("Dimensions names in paramter 'data' should match 'lat_dim' and 'lon_dim.")
  }
  if (all(names(dims) != c(lon_dim, lat_dim))) {
    data <- s2dv::Reorder(data, c(lon_dim, lat_dim))
  }
  if (length(lon) != dim(data)[lon_dim]) {
    stop("Length of parameter 'lon' should match dimension 'lon_dim' in 'data'.")
  }
  if (length(lat) != dim(data)[lat_dim]) {
    stop("Length of parameter 'lat' should match dimension 'lat_dim' in 'data'.")
  }
  if (!is.null(original_proj)) {
    original_proj <- st_crs(as.numeric(original_proj))
  } else {
    stop("Parameter 'original_proj' cannot be NULL.")
  }
  if (!is.null(target_proj)) {
    target_proj <- st_crs(as.numeric(target_proj))
  } else {
    stop("Parameter 'target_proj' cannot be NULL.")
  }
  if (is.unsorted(lon)) {
    warning("Parameter 'lon' has been tested for values from 0 to 359.")
  }
  sort_lat <- FALSE
  if (is.unsorted(rev(lat))) {
    lat <- rev(lat)
    data <- ClimProjDiags::Subset(data, along = lat_dim, indices = seq(length(lat), 1, -1))
    warning("Parameter 'lat' has been tested for values from -90 to 90 and 90 to 90.")
    sort_lat <- TRUE
  }
  if (length(brks) != length(cols) + 1) {
    stop("Parameter 'brks' should be more one than the number of colors in 'col'.")
  }
  ## Adapt ColorBar parameters to ggplot plot
  if (legend == 's2dv' && !is.null(col_inf) && !is.null(col_sup)) {
    max_brk <- max(brks)
    min_brk <- min(brks)
    step_brk <- abs(brks[2] - brks[1])
    data_inf <- min(data)
    data_sup <- max(data)
    data[which(data < min_brk)] <- min_brk - step_brk/2
    data[which(data > max_brk)] <- max_brk + step_brk/2
    brks_ggplot <- c(min_brk - step_brk, brks, max_brk + step_brk)
    cols_ggplot <- c(col_inf, cols, col_sup)
 } else if (legend == 's2dv' && !is.null(col_inf) && is.null(col_sup)) {
    min_brk <- min(brks)
    step_brk <- abs(brks[2] - brks[1])
    data_inf <- min(data)
    data[which(data < min_brk)] <- min_brk - step_brk/2
    #data[which(data > max_brk)] <- max_brk + step_brk/2
    brks_ggplot <- c(min_brk - step_brk, brks)
    cols_ggplot <- c(col_inf, cols)
  } else if (legend == 's2dv' && is.null(col_inf) && !is.null(col_sup)) {
    max_brk <- max(brks)
    step_brk <- abs(brks[2] - brks[1])
    data_sup <- max(data)
    #data[which(data < min_brk)] <- min_brk - step_brk/2
    data[which(data > max_brk)] <- max_brk + step_brk/2
    brks_ggplot <- c(brks, max_brk + step_brk)
    cols_ggplot <- c(cols, col_sup)
  } else {
    brks_ggplot <- brks
    cols_ggplot <- cols
  }
  data_df <- data.frame(lon = rep(as.vector(lon), length(lat)),
                        lat = sort(rep(as.vector(lat), length(lon)), 
                                decreasing = TRUE),
                        dat = as.vector(data))

  data_df <- st_as_sf(data_df, coords = c("lon", "lat"), crs = original_proj)
  data_df <- st_transform(data_df, crs = target_proj)

  data_df <- data_df %>%
  dplyr::mutate( long = st_coordinates(data_df)[,1],
                 lat = st_coordinates(data_df)[,2])
  # re-project dots
  if (!is.null(dots)) {
    if (all(names(dim(dots)) != c(lon_dim, lat_dim))) {
      dots <- s2dv::Reorder(dots, c(lon_dim, lat_dim))
    }
    if (sort_lat) {
      dots <- ClimProjDiags::Subset(dots, along = lat_dim, 
                                    indices = seq(length(lat), 1, -1))
    }
    dots_df <- data.frame(lon = rep(as.vector(lon), length(lat)),
                          lat = sort(rep(as.vector(lat), length(lon)),
                                    decreasing = TRUE),
                          dot = as.vector(dots))

    dots_df <- st_as_sf(dots_df, coords = c("lon", "lat"), crs = original_proj)
    dots_df <- st_transform(dots_df, crs = target_proj)

    dots_df <- dots_df %>%
    dplyr::mutate( long = st_coordinates(dots_df)[,1],
                   lat = st_coordinates(dots_df)[,2])
    dots_df <- subset(dots_df, dot == 0)
  }
  coastlines <- rnaturalearth::ne_coastline(scale = "medium", returnclass = "sf")
  coastlines <- st_transform(coastlines, crs = target_proj)
  p <- ggplot(data = data_df) + 
       geom_point(aes(x = long, y = lat, col = cut(dat, breaks = brks_ggplot)), size = size) +
       geom_sf(data = coastlines, colour ='black', size = coastlines_width) +
       xlab('') + ylab('')+ theme_bw()
  if (!is.null(dots)) {
    p <- p + geom_point(data = dots_df, aes(x = long, y = lat), 
                        shape = dots_shape, size = dots_size)
  }  
 
  if (is.null(legend) || legend == 's2dv') {
    p <- p +  scale_colour_manual(values = cols_ggplot, 
                                  drop = FALSE, na.value = colNA) + theme_void() +
      theme(legend.position = "none") 
  } else if (legend == 'ggplot2') {
      p <- p +  scale_colour_manual(values = cols_ggplot, 
                                    drop = FALSE, na.value = colNA) +
           guides(colour = guide_legend(override.aes = list(size = .2)))
  } else {
    stop("Parameter 'legend' must be NULL, ggplot2 or s2dv.")
  }
  if (!is.null(toptitle)) {
    p <- p + ggtitle(toptitle) +
         theme(plot.title = element_text(hjust = 0.5))
  }
  if (!is.null(caption)) {
    p <- p + labs(caption = caption) +
         theme(plot.caption = element_text(hjust = 0))
  }
  if (!is.null(fileout)) {
    if (!is.null(legend)) {
      if (legend == 's2dv') {
        if (vertical) {
          legend <- function() {
            par(mar = c(7.1, 2.2, 7.1, 3.1),
                mgp = c(3, 1, 0)
            )
            s2dv::ColorBar(brks = brks, cols, triangle_ends = triangle_ends, 
                           col_inf = col_inf, col_sup = col_sup,
                           vertical = vertical, ...)
          }
          p_legend <- cowplot::plot_grid(p, legend, rel_widths = c(6, 1))
        } else {
          legend <- function() {
            par(mar = c(1.1, 1.2, .1, 1.1),
                mgp = c(3, 1, 0)
            )
            s2dv::ColorBar(brks = brks, cols, triangle_ends = triangle_ends,
                           col_inf = col_inf, col_sup = col_sup,
                           vertical = vertical, ...)
          }
          p_legend <- cowplot::plot_grid(p, legend, rel_heights = c(6, 1),
                                         ncol = 1)
        }
        ggsave(fileout, p_legend, width = width, height = height, dpi = dpi, 
               units = units, device = device)
      }
    } else {
      ggsave(fileout, p, width = width, height = height, dpi = dpi, 
             units = units, device = device)
    }  
  } else {
    warning("Legend cannot be added in a pop-up window. Sat 'fileout' to add a legend.")
    p 
  } 
}

